#include "videocontainer.h"

#include <QApplication>
#include <QProgressDialog>
#include <QtConcurrent/QtConcurrent>
#include <QThread>

#include "opencv2/videoio/videoio.hpp"
#include "opencv2/imgcodecs.hpp"
#include "../globalobjects.h"


VideoContainer::VideoContainer(QObject *parent) :
    QObject(parent),
    fps(0),
    frameCount(0),
    duration(0)
{
    if(!tmpDir.isValid()) {
        qWarning() << "Could not create temporary folder!";
        exit(EXIT_FAILURE);
    }

    qDebug() << "Created tmp dir: " << tmpDir.path();


}

bool VideoContainer::loadFile(QString fileName)
{
    // copy file to temp dir
    QFile vid(fileName);
    videoFilePath = tmpDir.path()+"/"+QFileInfo(vid).fileName();

    if(!vid.copy(videoFilePath)) {
        qWarning() << "could not copy video";
        return false;
    }

    // create frames folder
    if(!QDir(tmpDir.path()).mkdir("frames")) {
        qWarning() << "could not create frames folder";
        return false;
    }



    // --- read properties of video ---
    cv::VideoCapture cap(videoFilePath.toStdString());

    // fps
    fps = cap.get(cv::CAP_PROP_FPS);
    qDebug() << "cv fps:" << fps;

    if(fps <= 0) {
        qWarning() << "invalid fps";
        return false;
    }

    // framecount
    frameCount = (int)cap.get(cv::CAP_PROP_FRAME_COUNT);
    qDebug() << "cv framecount:" << frameCount;

    if(frameCount <= 0) {
        qWarning() << "invalid framecount";
        return false;
    }

    // duration
    duration = double(frameCount) / fps;
    qDebug() << "cv duration:" << duration;

    // --- read video size ---
    vidSize.setHeight(cap.get(cv::CAP_PROP_FRAME_HEIGHT));
    vidSize.setWidth(cap.get(cv::CAP_PROP_FRAME_WIDTH));

    // --- read video frame by frame using multi threading ---
    int numberOfParallelThreads = int(QThread::idealThreadCount());
    if (numberOfParallelThreads==-1)
        numberOfParallelThreads=4;

    int frameSpace = round(frameCount/numberOfParallelThreads);

    for(int i=1;i<numberOfParallelThreads;i++)
        frameVector.append(i*frameSpace);
    frameVector.append(frameCount);

    // Create a progress dialog.
    QProgressDialog *dialog = new QProgressDialog(QString("Extracting %1 frames. Please wait.").arg(frameCount), "Abort",0,0);
    dialog->setWindowModality(Qt::WindowModal);

    // Create a QFutureWatcher and connect signals and slots.
    QFutureWatcher<void> futureWatcher;
    QObject::connect(&futureWatcher, SIGNAL(finished()), dialog, SLOT(reset()));
    QObject::connect(dialog, SIGNAL(canceled()), &futureWatcher, SLOT(cancel()));
    QObject::connect(dialog, SIGNAL(canceled()), this, SLOT(setCancel()));

    //create wrapper

    videoToFramesWrapper wrap(this);

    // Start the computation.
    futureWatcher.setFuture(QtConcurrent::map(frameVector, wrap));

    // Display the dialog and start the event loop.
    dialog->exec();
    futureWatcher.waitForFinished(); //block main application
    delete dialog;

    if (abortVideoToFrames){
        return false;
    }
    else{
    return true;
    }
}

// get the filepath+name of the image for a specific frame
QString VideoContainer::getFramePath(int frame)
{
    return QString(tmpDir.path()+"/frames/frame%1.jpg").arg(frame, 10, 10, QChar('0'));
}

void VideoContainer::setCancel()
{
    abortVideoToFrames=true;
}

void VideoContainer::videoToFrames(int &iteration)
{

    cv::VideoCapture cap(videoFilePath.toStdString());
    int index = frameVector.indexOf(iteration);
    int start = 0;

    if(index>0)
        start = frameVector.at(index-1);

    cap.set(cv::CAP_PROP_POS_FRAMES,start-1);
    for (int i = start ; i < iteration && !abortVideoToFrames; ++i) {
        cv::Mat image;
         cap.read(image);
        if(image.cols != 0 && image.rows != 0) {
            cv::imwrite(this->getFramePath(i).toUtf8().constData(), image);
        } else {
            qDebug() << "frame" << i << ": invalid image or no images left. skip.";
        }
    }
    //qDebug() << "iteration" << iteration << "in thread" << QThread::currentThreadId();
}
