#include "mainwindow.h"

#include <QDebug>
#include <QGraphicsScene>
#include <QMenu>
#include <QFileDialog>
#include <QApplication>
#include <QFile>
#include <QMessageBox>


#include "../globalobjects.h"
#include "../Features/morph.h"
#include "../Features/aoitracking.h"
#include "../Features/exportxml.h"
#include "../Features/aoisegmentation.h"

MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags) :
    QMainWindow(parent, flags),
    breakFrame(-1)
{

    qDebug() << "iSeg version:" << ISEG_VERSION;

    // ==== SETUP UI ====

    resize(1024,768);

    mainToolBar = new QToolBar("Toolbar");
    addToolBar(Qt::LeftToolBarArea, mainToolBar);
    mainToolBar->setIconSize(QSize(48,48));

    mainStatusBar = new QStatusBar();
    setStatusBar(mainStatusBar);

    setMenuBar(new QMenuBar());
    fileMenu = menuBar()->addMenu("&File");
    frameMenu = menuBar()->addMenu("F&rame");
    toolMenu = menuBar()->addMenu("&Tools");
    windowsMenu = menuBar()->addMenu("&Windows");
    helpMenu = menuBar()->addMenu("&Help");

    frameViewer = new FrameViewer(this);
    setCentralWidget(frameViewer);

    setDockNestingEnabled(true);

    frameNav = new FrameNav("Frame Navigation", this);
    frameNav->setAllowedAreas(Qt::AllDockWidgetAreas);
    addDockWidget(Qt::TopDockWidgetArea, frameNav);

    timeLine = new TimeLine("Time Line", this);
    timeLine->setAllowedAreas(Qt::AllDockWidgetAreas);
    addDockWidget(Qt::BottomDockWidgetArea, timeLine, Qt::Vertical);



    // ==== SETUP UI ACTIONS ====

    windowsMenu->addAction(frameNav->toggleViewAction());
    windowsMenu->addAction(timeLine->toggleViewAction());



    // ==== SETUP LOGIC ACTIONS ====

    actionNew = new QAction("&New", this);
    actionNew->setShortcut(QKeySequence::New);
    actionNew->setIcon(QIcon(":icons/new.ico"));
    fileMenu->addAction(actionNew);
    mainToolBar->addAction(actionNew);
    connect(actionNew, SIGNAL(triggered()), this, SLOT(onActionNew()));

    actionSave = new QAction("&Save", this);
    actionSave->setShortcut(QKeySequence::Save);
    actionSave->setIcon(QIcon(":icons/save.ico"));
    fileMenu->addAction(actionSave);
    mainToolBar->addAction(actionSave);
    actionSave->setEnabled(false);
    connect(actionSave, SIGNAL(triggered()), this, SLOT(onActionSave()));

    actionOpen = new QAction("&Open", this);
    actionOpen->setShortcut(QKeySequence::Open);
    actionOpen->setIcon(QIcon(":icons/open.ico"));
    fileMenu->addAction(actionOpen);
    mainToolBar->addAction(actionOpen);
    connect(actionOpen, SIGNAL(triggered()), this, SLOT(onActionOpen()));

    actionClose = new QAction("&Close", this);
    actionClose->setShortcuts(QList<QKeySequence>() << QKeySequence::Quit << QKeySequence::Close);
    fileMenu->addSeparator();
    fileMenu->addAction(actionClose);
    connect(actionClose, SIGNAL(triggered()), this, SLOT(onActionClose()));


    actionNextFrame = new QAction("Next Frame", this);
    actionNextFrame->setIcon(QIcon(":icons/goNext.ico"));
    actionNextFrame->setShortcut(QKeySequence(Qt::Key_Right));
    frameMenu->addAction(actionNextFrame);
    connect(actionNextFrame, SIGNAL(triggered()), this, SLOT(onActionNextFrame()));

    actionPrevFrame = new QAction("Previous Frame", this);
    actionPrevFrame->setIcon(QIcon(":icons/goPre.ico"));
    actionPrevFrame->setShortcut(QKeySequence(Qt::Key_Left));
    frameMenu->addAction(actionPrevFrame);
    connect(actionPrevFrame, SIGNAL(triggered()), this, SLOT(onActionPrevFrame()));


    actionLinearMorphing = new QAction("&morph", this);
    actionLinearMorphing->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_M));
    actionLinearMorphing->setIcon(QIcon(":icons/morph.ico"));
    toolMenu->addAction(actionLinearMorphing);
    mainToolBar->addAction(actionLinearMorphing);
    actionLinearMorphing->setEnabled(false);
    connect(actionLinearMorphing, SIGNAL(triggered()), this, SLOT(onActionLinearMorphing()));

    actionInteractiveTracking = new QAction("&iTrack", this);
    actionInteractiveTracking->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_I));
    actionInteractiveTracking->setIcon(QIcon(":icons/iTrack.ico"));
    toolMenu->addAction(actionInteractiveTracking);
    mainToolBar->addAction(actionInteractiveTracking);
    actionInteractiveTracking->setEnabled(false);
    connect(actionInteractiveTracking, SIGNAL(triggered()), this, SLOT(onActionInteractiveTracking()));

    actionAutomaticTracking = new QAction("&aTrack", this);
    actionAutomaticTracking->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_A));
    actionAutomaticTracking->setIcon(QIcon(":icons/aTrack.ico"));
    toolMenu->addAction(actionAutomaticTracking);
    mainToolBar->addAction(actionAutomaticTracking);
    actionAutomaticTracking->setEnabled(false);
    connect(actionAutomaticTracking, SIGNAL(triggered()), this, SLOT(onActionAutomaticTracking()));

    actionExportXML= new QAction("export XML", this);
    actionExportXML->setIcon(QIcon(":icons/export.ico"));
    toolMenu->addAction(actionExportXML);
    mainToolBar->addAction(actionExportXML);
    actionExportXML->setEnabled(false);
    connect(actionExportXML, SIGNAL(triggered()),this, SLOT(onActionExportXML()));

    actionSegmentation = new QAction("segmentation", this);
    actionSegmentation->setIcon(QIcon(":icons/scissors.ico"));
    toolMenu->addAction(actionSegmentation);
    mainToolBar->addAction(actionSegmentation);
    actionSegmentation->setEnabled(false);
    connect(actionSegmentation, SIGNAL(triggered()),this, SLOT(onActionSegmentation()));

    helpLicense=new QAction("License Notes",this);
    helpMenu->addAction(helpLicense);
    connect(helpLicense,SIGNAL(triggered()),this,SLOT(onActionHelpLicense()));

    helpAbout=new QAction("About iSeg",this);
    helpMenu->addAction(helpAbout);
    connect(helpAbout,SIGNAL(triggered()),this,SLOT(onActionHelpAbout()));

}

MainWindow::~MainWindow()
{

}

void MainWindow::recreateDataState()
{
    // create new state
    delete g_state;
    g_state=NULL;
    g_state = new DataState();

    frameNav->reload();
    frameViewer->reload();
    timeLine->reload();

    setDisabled(false);
    reload();

    // connect ui to state
    frameNav->connectToState();
    frameViewer->connectToState();
    timeLine->connectToState();

    return;
}

void MainWindow::reload()
{
    if(!g_state->videoLoaded){
        actionSave->setEnabled(false);
        actionLinearMorphing->setEnabled(false);
        actionInteractiveTracking->setEnabled(false);
        actionAutomaticTracking->setEnabled(false);
        actionExportXML->setEnabled(false);
        actionSegmentation->setEnabled(false);
    }
    else{
        frameViewer->reload();
        frameNav->reload();
        timeLine->reload();

        actionSave->setEnabled(true);

        if(g_state->hasObjects()){
            actionLinearMorphing->setEnabled(true);
            actionInteractiveTracking->setEnabled(true);
            actionAutomaticTracking->setEnabled(true);
            actionExportXML->setEnabled(true);
            actionSegmentation->setEnabled(true);
        }
        else{
            actionLinearMorphing->setEnabled(false);
            actionInteractiveTracking->setEnabled(false);
            actionAutomaticTracking->setEnabled(false);
            actionExportXML->setEnabled(false);
            actionSegmentation->setEnabled(false);
        }

    }
    return;
}

void MainWindow::onActionNew()
{
    setDisabled(true);

    // TODO: make only video files selctable in the qfiledialog
    QString fileName = QFileDialog::getOpenFileName(this, "Video file", QDir::homePath(),  tr("supported file types (*.avi *.mpeg *.mp4 *.m4a *.m4p *.m4b *.m4r *.m4v *.wmv *.h263);;AVI (*.avi);;MPEG (*.mpeg);;MPEG4 (*.mp4 *.m4a *.m4p *.m4b *.m4r *.m4v);;WMV (*.wmv);;H.263 (*.h263) " ), 0, QFileDialog::DontUseNativeDialog);
    QApplication::processEvents();
    if(fileName.isEmpty()) {
        setDisabled(false);
        reload();
        return;
    }
    if(g_state->videoLoaded) {
        recreateDataState();
    }

    qDebug() << "Load file " << fileName;
    qDebug() << "Start processing...";

    g_state->vid = new VideoContainer(this);
    if(g_state->vid->loadFile(fileName)){
        g_state->videoLoaded = true;
        g_state->currentFrame = 0;
        setDisabled(false);
        reload();
        return;
    }
    else{
        recreateDataState();
        return;
    }
}

void MainWindow::onActionSave()
{
    if(g_state->videoLoaded) {
        QString fileName = QFileDialog::getSaveFileName(this,tr("Save File"), QDir::homePath(),tr("iSeg Data (*.iseg);;All Files (*.iseg)"));
        if (fileName.isEmpty())
            return;
        else {
            QFileInfo fi(fileName);
            if(fi.suffix()!="iseg")
                fileName=fileName+".iseg";
            QFile file(fileName);
            if(!file.open(QIODevice::WriteOnly)){
                qDebug() <<"Could not open file location";
                // TODO: added user warning
                return;
            }

            QDataStream out(&file);
            out.setVersion(QDataStream::Qt_5_3);
            out<<g_state->vid->videoFilePath.split("/").last()<<QString(ISEG_VERSION)<<g_state->vid->frameCount<<g_state->vid->vidSize<<*g_state;
            file.flush();
            file.close();
        }
    }
    return;
}

void MainWindow::onActionOpen()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Select iseg data file"), QDir::homePath(),tr("iSeg Data (*.iseg);;All Files (*.iseg)"), 0, QFileDialog::DontUseNativeDialog);
    QApplication::processEvents();
    if(fileName == "") {
        setDisabled(false);
        return;
    }

    QString videoFile="error.error";
    QString tempDataStateVersion = "";


    QFile file(fileName);
    if(!file.open(QIODevice::ReadOnly)){
       qDebug() <<"Could not open file";
        return;
    }
    QDataStream in(&file);
    in.setVersion(QDataStream::Qt_5_3);
    in >>videoFile;

    // TODO: make only video files selctable in the qfiledialog
    QString filter = QString("Video ("+videoFile.replace( " ", "*" )+");;All files (*)");
    QString videoFileName = QFileDialog::getOpenFileName(this, tr("Select video file"),"", filter, 0, QFileDialog::DontUseNativeDialog);
    QApplication::processEvents();
    if(videoFileName == "") {
        setDisabled(false);
        return;
    }

    if(g_state->videoLoaded) {
        recreateDataState();
    }

    qDebug() << "Load file " << videoFileName;
    qDebug() << "Start processing...";

    g_state->vid = new VideoContainer(this);
    if(!g_state->vid->loadFile(videoFileName)){
        recreateDataState();
        return;
    }
    g_state->videoLoaded = true;

    int tempFrameCount=0;
    QSize tempVidSize(0,0);

    in >>tempDataStateVersion>>tempFrameCount>>tempVidSize>>*g_state;
    file.close();
    qDebug()<<g_state->getCurrentObject()->getFrameSpans();
    if(tempDataStateVersion!=ISEG_VERSION || tempFrameCount!=g_state->vid->frameCount || tempVidSize!=g_state->vid->vidSize){
        QMessageBox::information(this,"iSeg", tr("iSeg file does not match the requierments (e.g. version of iSeg, video frame rate or video resolution)"));
        recreateDataState();
        reload();
        return;
    }

    setDisabled(false);
    reload();
    return;
}

void MainWindow::onActionClose()
{
    if(g_state->videoLoaded) {
        // TODO - unsaved project close
        qWarning() << "already loaded. TODO";
        close();

        return;
    }
    close();
}

void MainWindow::onActionNextFrame()
{
    if(!g_state->videoLoaded) return;
    g_state->nextFrame();
}

void MainWindow::onActionPrevFrame()
{
    if(!g_state->videoLoaded) return;
    g_state->prevFrame();
}

void MainWindow::onActionLinearMorphing()
{

    Morph doMorphing(g_state);
    doMorphing.linearMorphing();
}

void MainWindow::onActionInteractiveTracking()
{
    AOItracking* tracking = new AOItracking(g_state,true);
    if(breakFrame!=-1)
        onActionLinearMorphing();

    breakFrame=tracking->interactiveTracking(breakFrame);
    if(breakFrame!=-1){
        g_state->getCurrentObject()->keyFrames.at(breakFrame)->reducePolygonVertices();
        g_state->getCurrentObject()->keyFrames.at(breakFrame)->setType(KeyFrame::SetVisible);
        g_state->setFrame(breakFrame);
        mainStatusBar->showMessage(tr("User interaction required!"),2000);

    }
    else{
        mainStatusBar->showMessage(tr("Interactive AOI tracking complete."),2000);
        QMessageBox msgBox;
        msgBox.setText("Interactive AOI tracking complete.");
        msgBox.setIcon(QMessageBox::Information);
        msgBox.exec();
    }

    delete tracking;
}

void MainWindow::onActionAutomaticTracking()
{
    AOItracking* tracking = new AOItracking(g_state);
    tracking->automaticTracking();
    mainStatusBar->showMessage(tr("Automatic AOI tracking complete."),2000);
    QMessageBox msgBox;
    msgBox.setText("Automatic AOI tracking complete.");
    msgBox.setIcon(QMessageBox::Information);
    msgBox.exec();
    delete tracking;
}

void MainWindow::onActionExportXML()
{
    QString selectedFilter;
    QString fileName = QFileDialog::getSaveFileName(this,
                                tr("QFileDialog::getSaveFileName()"),
                                "",
                                tr("Polygon lables - ViPER like (*.xml);;ViPER Bounding Box (*.xml )"),
                                &selectedFilter);
    if (fileName.isEmpty() || selectedFilter.isEmpty())
        return;
    else {
        QFileInfo fi(fileName);
        if(fi.suffix()!="xml")
            fileName=fileName+".xml";

        QFile *file = new QFile(fileName);
        ExportXML* xml;
        if(selectedFilter.contains("ViPER Bounding Box")){
            xml=new ExportXML(g_state,file,true);
        }
        else{
            xml=new ExportXML(g_state,file);
        }

        if(!xml->saveXML()){
            QMessageBox msgBox;
            msgBox.setText("Could not write .xml file.");
            msgBox.setInformativeText("Do you want to retry saving?");
            msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
            msgBox.setDefaultButton(QMessageBox::Yes);
            int ret = msgBox.exec();
            if(ret==QMessageBox::Yes)
                onActionExportXML();
        }
        delete file;
    }
}

void MainWindow::onActionSegmentation()
{
    AoiSegmentation aoiSegment(g_state);
    aoiSegment.runSegmenation();
}


void MainWindow::onActionTmpDebugAction()
{
    qDebug() << "DEBUG ACTION TRIGGERED WITH PRESSING 'CTRL+D'";
    g_state->getCurrentObject()->getCurrentKeyFrame()->reducePolygonVertices();
}

void MainWindow::onActionHelpLicense()
{
    QMessageBox msgBox;
    QString str = QString("<h3><u>i</u>nteractive <u>Se</u>mi-automatic <u>g</u>round truth annotation tool - <strong>iSeg</strong></h3><br>"
                          "<strong>Version:<strong> %1 alpha<br><br><br>"
                          "iSeg tool and its source:<ul><li> <a href=\"http://www.gnu.org/licenses/gpl-3.0.en.html\">GPLv3</a></li></ul>"
                          " This program is free software: you can redistribute it and/or modify"
                          "it under the terms of the GNU General Public License as published by"
                          "the Free Software Foundation, either version 3 of the License, or"
                          "(at your option) any later version."

                          "This program is distributed in the hope that it will be useful,"
                          "but WITHOUT ANY WARRANTY; without even the implied warranty of"
                          "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
                          "GNU General Public License for more details."

                          "You should have received a copy of the GNU General Public License"
                          "along with this program.  If not, see <a href=\"http://www.gnu.org/licenses/\">http://www.gnu.org/licenses/</a>."
                          "Third-party<ul>"
                          "<li>QT: <a href=\"http://doc.qt.io/qt-5/lgpl.html\">LGPLv3</a></li>"
                          "<li>OpenCV: <a href=\"http://opencv.org/license.html\">BSD</a></li>"
                          "<li>Eigen: <a href=\"http://eigen.tuxfamily.org/index.php?title=Licensing_FAQ\">MPL2</a></li></ul><br>"
                          "By downloading, copying, installing or using the software you agree to this licenses."
                          "<br><br>For more information, see Help/About"
                          ).arg(g_state->dataStateVersion);
    msgBox.about(this, "Licenses iSeg",str);
}


void MainWindow::onActionHelpAbout()
{
    QMessageBox msgBox;
    QString str = QString("<h3><u>i</u>nteractive <u>Se</u>mi-automatic <u>g</u>round truth annotation tool - <strong>iSeg</strong></h3><br>"
                          "<strong>Version:<strong> %1 <br><br>"
                          "If you use iSeg please cite:<br> <i>Julius Schöning, Patrick Faion, and Gunther Heidemann. 2015. Semi-automatic ground truth annotation in videos: An interactive tool for polygon-based object annotation and segmentation. In Proceedings of the 8th International Conference on Knowledge Capture (K-CAP 2015). ACM, New York, NY, USA,<br><a href=\"http://doi.acm.org/10.1145/2815833.2816947\">http://doi.acm.org/10.1145/2815833.2816947</a></i>"
                          "<br><br>For more information, see <a href=\"https://ikw.uos.de/~cv/projects/iSeg\">https://ikw.uos.de/~cv/projects/iSeg</a>"
                          "<br><br>2016 Julius Schöning and Paktrick Faion."
                          ).arg(g_state->dataStateVersion);
    msgBox.about(this, "About: iSeg",str);
}




